""" dialog to edit docType schema """
from enum import Enum
from typing import Any
import numpy as np
import pandas as pd
from PySide6.QtCore import Slot
from PySide6.QtWidgets import (QComboBox, QDialog, QDialogButtonBox, QInputDialog, QMessageBox, QTabBar, QTableWidget,
                               QTableWidgetItem, QTabWidget, QVBoxLayout)
from ...fixedStringsJson import defaultDataHierarchyNode
from ..guiCommunicate import Communicate
from ..guiStyle import IconButton, Label, TextButton, widgetAndLayout
from ..messageDialog import showMessage
from .delete_column_delegate import DeleteColumnDelegate
from .docTypeEdit import DocTypeEditor
from .listFreeDelegate import ListFreeDelegate
from .listItemDelegate import ListItemDelegate
from .mandatory_column_delegate import MandatoryColumnDelegate
from .name_column_delegate import NameColumnDelegate
from .reorder_column_delegate import ReorderColumnDelegate

#                0       1            2      3           4      5         6
COLUMN_NAMES = ['name','description','unit','mandatory','item list','free list','move up','delete']
COLUMN_WIDTH = [100,   250,          60,    80,         150,        200,        60,       50]
pd.options.mode.copy_on_write = True

class SchemeEditor(QDialog):
  """ dialog to edit docType schema """
  def __init__(self, comm:Communicate):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
      doc (dict):  document to change / create
    """
    super().__init__()
    self.comm = comm
    self.comm.backendThread.worker.beSendSQL.connect(self.onGetData)
    self.docType = ''
    self.docLabel= ''
    self.cmd     = ''
    self.df      = pd.DataFrame()
    self.docTypesLabels = [(k,v['title']) for k,v in self.comm.docTypesTitles.items()]
    self.closeButtons:list[IconButton] = []                                             #close buttons of tabs
    self.setWindowTitle('Item type editor')
    self.restartAfterClose = False

    # GUI elements
    self.setMinimumWidth(int(np.sum(COLUMN_WIDTH))+80)
    self.setMinimumHeight(500)
    mainL = QVBoxLayout(self)
    Label('Item type editor', 'h1', mainL)
    Label('Warning: every change of the item type saves that content', 'h3', mainL)
    _, docTypeL = widgetAndLayout('H', mainL, 's')
    Label('Item type:', '', docTypeL)
    self.selectDocType = QComboBox()
    self.selectDocType.currentTextChanged.connect(self.changeDocType)
    self.selectDocType.setStyleSheet(self.comm.palette.get('secondaryText','color'))
    docTypeL.addWidget(self.selectDocType)
    docTypeL.addStretch(1)
    IconButton('fa5s.plus',  self, [Command.NEW],  docTypeL, tooltip='New item type')
    IconButton('fa5s.edit',  self, [Command.EDIT], docTypeL, tooltip='Edit item type')
    IconButton('fa5s.trash', self, [Command.DEL],  docTypeL, tooltip='Delete item type')

    # tabs: empty
    self.tabW = QTabWidget(self)
    self.tabW.tabBarDoubleClicked.connect(self.renameTab)
    self.tabW.tabBarClicked.connect(self.createNewTab)
    mainL.addWidget(self.tabW, stretch=10)
    self.nameColumnDelegates     :list[NameColumnDelegate]      = []
    self.mandatoryColumnDelegates:list[MandatoryColumnDelegate] = []
    self.listItemDelegates       :list[ListItemDelegate]        = []
    self.listFreeDelegates       :list[ListFreeDelegate]        = []
    self.reorderColumnDelegates  :list[ReorderColumnDelegate]   = []
    self.deleteColumnDelegates   :list[DeleteColumnDelegate]    = []
    self.newWidget = QTableWidget()
    #final button box
    mainL.addStretch(1)
    buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Save | QDialogButtonBox.StandardButton.Cancel)
    buttonBox.clicked.connect(self.closeDialog)
    mainL.addWidget(buttonBox)
    #initialize
    _ = [self.selectDocType.addItem(v,k) for k,v in self.docTypesLabels]
    self.selectDocType.setCurrentIndex(0)
    return


  @Slot(str, pd.DataFrame)
  def onGetData(self, cmd:str, data:pd.DataFrame) -> None:
    """ Handle data received from backend worker
    Args:
      cmd (str): command that was sent
      data (pd.DataFrame): DataFrame containing the data
    """
    if cmd==self.cmd:
      self.df = data
      self.paint()


  def changeDocType(self, label:str) -> None:
    """ Change the item / document type

    Args:
      label (str): label of item / document type
    """
    # get data frame
    if not label:
      return
    if self.docType:
      self.finishDocType()
    self.docLabel= label
    fittingDocTypes = [k for k,v in self.docTypesLabels if v==label]
    if not fittingDocTypes:                                        #nothing to save since docType just created
      self.df = pd.DataFrame(defaultDataHierarchyNode)
      self.df['docType'] = self.selectDocType.currentData()
      self.df.drop(['PURL'], axis=1, inplace=True)
      self.df.rename(columns={'query':'long'}, inplace=True)
      self.paint()
      return
    self.docType = fittingDocTypes[0]
    self.cmd = 'SELECT docTypeSchema.docType, docTypeSchema.class, docTypeSchema.idx, docTypeSchema.name, '\
          'docTypeSchema.unit, docTypeSchema.mandatory, docTypeSchema.list, definitions.long '\
          'FROM docTypeSchema LEFT JOIN definitions ON definitions.key = (docTypeSchema.class || "." || docTypeSchema.name) '\
          f'WHERE docTypeSchema.docType=="{self.docType}"'
    self.comm.uiSendSQL.emit([{'type':'get_df', 'cmd':self.cmd}])

  def paint(self) -> None:
    """ Paint the dialog with the data """
    df = self.df.rename(columns={'long':'description'})
    df['item list'] = df['list'].apply(lambda x: '' if ',' in x else x)
    df['free list'] = df['list'].apply(lambda x: x  if ',' in x else '')
    df = df.drop('list', axis=1)
    df['idx'] = df['idx'].astype('int')
    # GUI
    self.tabW.clear()
    self.reorderColumnDelegates.clear()
    for group in df['class'].unique():
      data  = df[df['class']==group]
      table = QTableWidget(len(data)+1, 8)
      table.verticalHeader().hide()
      table.setAlternatingRowColors(True)
      table.setHorizontalHeaderLabels(COLUMN_NAMES)
      table.horizontalHeader().setStyleSheet('QHeaderView::section {padding: 1px; margin: 0px;}')
      for idx, width in enumerate(COLUMN_WIDTH[:-1]):
        table.setColumnWidth(idx, width)
      for _,row in data.iterrows():
        idx = int(row['idx'])
        for col in range(6):
          table.setItem(idx, col, QTableWidgetItem(row[COLUMN_NAMES[col]]))
      #Delegates to give function to row
      self.nameColumnDelegates.append(NameColumnDelegate())
      table.setItemDelegateForColumn(0, self.nameColumnDelegates[-1])
      self.mandatoryColumnDelegates.append(MandatoryColumnDelegate())
      table.setItemDelegateForColumn(3, self.mandatoryColumnDelegates[-1])
      self.listItemDelegates.append(ListItemDelegate(self.docTypesLabels))
      table.setItemDelegateForColumn(4, self.listItemDelegates[-1])
      self.listFreeDelegates.append(ListFreeDelegate())
      table.setItemDelegateForColumn(5, self.listFreeDelegates[-1])
      self.reorderColumnDelegates.append(ReorderColumnDelegate())
      table.setItemDelegateForColumn(6, self.reorderColumnDelegates[-1])
      self.deleteColumnDelegates.append(DeleteColumnDelegate())
      table.setItemDelegateForColumn(7, self.deleteColumnDelegates[-1])
      self.tabW.addTab(table, group or 'default')
    # define close buttons on some of the tabs
    self.closeButtons.clear()
    for idx in range(1, self.tabW.count()):
      self.closeButtons.append(IconButton('fa5s.times', self, [Command.DEL_GROUP,idx], None, 'Delete group'))
      self.tabW.tabBar().setTabButton(idx, QTabBar.ButtonPosition.RightSide, self.closeButtons[-1])
      header = table.horizontalHeader()
      header.setStretchLastSection(True)
    self.tabW.addTab(self.newWidget, '+')
    return


  def finishDocType(self) -> None:
    """
    Finish a docType and save to DB,
    - either by selecting another docType or by pressing save
    """
    df = self.table2schema()
    # verification: uniqueness in names. etc
    unique = ( df.groupby('class')['name'].nunique() == df.groupby('class')['name'].count()  ).all()
    if not unique:
      showMessage(self, 'Error', 'Within each table, the text in the first column has to be unique. E.g. no '
                  'two "tags" are allowed.', 'Critical')
      return
    # SAVE DATA
    dfSchema = df.drop('description', axis=1)
    dfDef = df[['class','name','description']]
    dfDef['key'] = dfDef['class']+'.'+dfDef['name']
    dfDef = dfDef.drop(['name','class'], axis=1)[['key','description']]
    dfDef.rename(columns={'description':'long'}, inplace=True)
    nonUnique = dict(dfDef.groupby(['key'])['long'].apply(lambda x: len(np.unique(x))))
    if nonUniqueStr:= ', '.join(k for k,v in nonUnique.items() if v>1):
      button = QMessageBox.question(self, 'Non unique definitions',
                                  f'The definitions are non-unique for {nonUniqueStr}. Do you want to flatten?',
                                  QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
      if button == QMessageBox.StandardButton.No:
        return
    dfDef = dfDef.groupby(['key']).first()
    listDef = list(dfDef.itertuples(name=None))
    # save to sqlite
    self.comm.uiSendSQL.emit([
      {'type':'many', 'cmd':'INSERT INTO definitions (key, long, PURL) VALUES (?, ?, "") ON CONFLICT(key) DO '
                            'UPDATE SET long = excluded.long;', 'list':listDef},
      {'type':'one',  'cmd':f"DELETE FROM docTypeSchema WHERE docType == '{self.docType}'"},
      {'type':'df',   'table':'docTypeSchema', 'df':dfSchema}])
    return


  def closeDialog(self, btn:TextButton) -> None:
    """
    cancel or save entered data

    Args:
      btn (QButton): save or cancel button
    """
    if btn.text().endswith('Cancel'):
      self.reject()
    elif btn.text().endswith('Save'):
      self.finishDocType()
      if self.restartAfterClose:
        self.comm.start()
      self.accept()
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.NEW:
      dialog = DocTypeEditor(self.comm, '', self.createdNewDocType)
      dialog.exec()
    elif command[0] is Command.EDIT:
      dialog = DocTypeEditor(self.comm, self.selectDocType.currentData())
      dialog.exec()
    elif command[0] is Command.DEL:
      button = QMessageBox.question(self, 'Question', 'Do you really want to remove the doc-type?',
                                    QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
      if button == QMessageBox.StandardButton.No:
        return
      self.comm.uiSendSQL.emit([
        {'type':'one', 'cmd':f"DELETE FROM docTypes WHERE docType == '{self.selectDocType.currentData()}'"},
        {'type':'one', 'cmd':f"DELETE FROM docTypeSchema WHERE docType == '{self.selectDocType.currentData()}'"}
      ])
      self.comm.commSendConfiguration.emit(self.comm.configuration, self.comm.projectGroup)#reset docTypes, etc.
      self.selectDocType.clear()
    elif command[0] is Command.DEL_GROUP:
      docLabel = str(self.docLabel)
      button = QMessageBox.question(self, 'Question', 'Do you really want to remove this group?',
                                    QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
      if button == QMessageBox.StandardButton.Yes:
        group = self.tabW.tabBar().tabText(command[1])
        self.comm.uiSendSQL.emit([{'type':'one','cmd':
          f"DELETE FROM docTypeSchema WHERE docType == '{self.docType}' AND class == '{group}'"}])
        self.changeDocType(docLabel)
    return


  def table2schema(self) -> pd.DataFrame:
    """ save text content of tables """
    df = pd.DataFrame()
    lastRow = 0
    for tabID in range(self.tabW.count()-1):                                           # -1: ignore last tab +
      group = self.tabW.tabBar().tabText(tabID)
      group = '' if group=='default' else group
      widget:QTableWidget = self.tabW.widget(tabID)                                 # type: ignore[assignment]
      for row in range(widget.rowCount()-1):                   # -1: ignore last row as it is meant for adding
        df.at[lastRow, 'docType'] = self.docType
        df.at[lastRow, 'class']   = group
        df.at[lastRow, 'idx']     = str(row)
        for col in range(6):
          item = widget.item(row,col)
          df.at[lastRow, COLUMN_NAMES[col]] = '' if item is None else item.text()
        lastRow += 1
    df['list'] = df['item list']+df['free list']
    df.drop(['item list','free list'], axis=1, inplace=True)
    return df


  def renameTab(self, idx:int) -> None:
    """ rename tab after double clicking

    Args:
      idx (int): index of the tab
    """
    if 0 < idx <self.tabW.count()-1:
      textOld = self.tabW.tabText(idx)
      textNew, ok = QInputDialog.getText(self, 'Rename group', 'Enter new group name:', text=textOld)
      if ok and textNew.strip():
        self.tabW.setTabText(idx, textNew.strip())
    return


  def createNewTab(self, idx:int) -> None:
    """ Create new group

    Args:
      idx (int): index of the tab
    """
    if idx == self.tabW.count()-1:
      textNew, ok = QInputDialog.getText(self, 'Create group', 'Enter new group name:', text='')
      if ok and textNew.strip():
        self.comm.uiSendSQL.emit([{'type':'one',
                                   'cmd':f"INSERT INTO docTypeSchema VALUES ({', '.join(['?']*7)})",
                                   'list':[self.docType, textNew.strip(), '0', 'item', '', '', '']}])
        self.docType = ''                               # prevent self.finishDocType from deleting new entries
        self.changeDocType(self.docLabel)
    return


  def createdNewDocType(self,label:str, docType:str) -> None:
    """ Callback function to change docType after creating a new one

    Args:
      label (str): label of the new docType
      docType (str): name of the new docType
    """
    if docType not in [self.selectDocType.itemData(i) for i in range(self.selectDocType.count())]:
      self.selectDocType.addItem(label, docType)
    self.restartAfterClose = True
    self.selectDocType.setCurrentText(label)
    return

  def reject(self) -> None:
    """ Reject the dialog, stop the thread and disconnect signals """
    self.comm.backendThread.worker.beSendSQL.disconnect(self.onGetData)
    super().reject()


  def accept(self) -> None:
    """ Accept the dialog, stop the thread and disconnect signals """
    self.comm.backendThread.worker.beSendSQL.disconnect(self.onGetData)
    super().accept()


class Command(Enum):
  """ Commands used in this file """
  NEW          = 1
  DEL          = 2
  EDIT         = 3
  DEL_GROUP    = 4



""" Entire config dialog (dialog is blocking the main-window, as opposed to create a new widget-window)"""
from PySide6.QtWidgets import QDialog, QTabWidget, QVBoxLayout, QWidget
from ..guiCommunicate import Communicate
from ..repositories.config import ConfigurationRepositories
from .addOnParameter import ConfigurationAddOnParameter
from .authors import ConfigurationAuthors
from .gui import ConfigurationGUI
from .projectGroup import ProjectGroup
from .setup import ConfigurationSetup

# Loading times of tabs: min - max of 3 runs
# Tab 'Project group' loaded in 0.022 - 0.023 seconds
# Tab 'Appearance'    loaded in 0.045 - 0.047 seconds
# Tab 'Author'        loaded in 0.014 seconds
# Tab 'Repository'    loaded in 0.013 seconds
# Tab 'Add-on parameters' loaded in 0.672 - 0.672 seconds
# Tab 'Setup'         loaded in 0.002 seconds


class Configuration(QDialog):
  """ Main class of entire config dialog """
  def __init__(self, comm:Communicate, startTab:str=''):
    """
    Initialization
    - if sidebar.py notices no database, it will call this dialog with 'setup' as only tab

    Args:
      backend (Pasta): backend, not communication
      startTab (str): tab to show initially
    """
    super().__init__()
    self.comm = comm
    self.setWindowTitle('PASTA-ELN configuration')
    self.setMinimumWidth(1200)

    # GUI elements
    mainL = QVBoxLayout(self)
    self.tabW = QTabWidget(self)
    self.tabW.currentChanged.connect(self.loadTab)
    mainL.addWidget(self.tabW)
    self.tabs:dict[str,QWidget] = {}                                        # to hold the actual tab instances
    self.placeholders = {}                                                       # to hold placeholder widgets
    self.tabClasses = {
      'Project group': ProjectGroup,
      'Interface': ConfigurationGUI,
      'Author': ConfigurationAuthors,
      'Repository': ConfigurationRepositories,
      'Add-on parameters': ConfigurationAddOnParameter,
      'Setup': ConfigurationSetup
    }
    for name in self.tabClasses:
      placeholder = QWidget()                                   # Only create empty placeholder tabs initially
      self.placeholders[name] = placeholder
      self.tabW.addTab(placeholder, name)

    if startTab=='setup':
      self.tabW.setCurrentIndex(self.tabW.indexOf(self.placeholders['Setup']))
      for tabName in self.tabClasses:
        if tabName != 'Setup':
          self.tabW.setTabEnabled(self.tabW.indexOf(self.placeholders[tabName]), False)
    else:
      self.tabW.setCurrentIndex(0)


  def loadTab(self, index:int) -> None:
    """ Load the tab content when the tab is clicked
    - Replace the placeholder's content (layout and widget inside it)

    Args:
      index (int): index of the tab that was clicked
    """
    tabName = self.tabW.tabText(index)
    if tabName not in self.tabs and tabName in self.tabClasses:
      # Create the actual tab content
      tabClass = self.tabClasses[tabName]
      self.tabs[tabName] = tabClass(self.comm, self.closeWidget)
      placeholder = self.placeholders[tabName]
      placeholderL = placeholder.layout()
      if placeholderL is not None:
        placeholderL.deleteLater()                                            # Clear the placeholder's layout
      layout = QVBoxLayout(placeholder)
      layout.setContentsMargins(0, 0, 0, 0)
      layout.addWidget(self.tabs[tabName])                                 # Add the actual tab content widget


  def closeWidget(self, restart:bool=True) -> None:
    """
    callback function to close widget, without restarting the entire app

    Args:
      restart (bool): if true, initialize
    """
    self.close()
    if restart:
      self.comm.start()
    return



#!/usr/bin/python3
"""TEST the form """
import logging, warnings, shutil
from pathlib import Path
from PySide6.QtCore import QEventLoop, Slot
from pasta_eln.backendWorker.backend import Backend
from pasta_eln.backendWorker.worker import Task
from pasta_eln.installationTools import exampleData
from pasta_eln.miscTools import getConfiguration
from pasta_eln.UI.form import Form
from pasta_eln.UI.guiCommunicate import Communicate
from pasta_eln.UI.palette import Palette
from .test_03_dragDrop import verify

def test_simple(qtbot, caplog):
  """
  main function
  """
  # initialization: create database, destroy on filesystem and database and then create new one
  warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
  warnings.filterwarnings('ignore', message='invalid escape sequence')
  warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
  warnings.filterwarnings('ignore', category=ImportWarning)
  logPath = Path.home()/'pastaELN.log'
  logging.basicConfig(filename=logPath, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                      datefmt='%m-%d %H:%M:%S')   #This logging is always info, since for installation only
  for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
    logging.getLogger(package).setLevel(logging.WARNING)
  logging.info('Start 10 test')

  # start app and create project
  configuration, _ = getConfiguration('research')
  exampleData(True, None, 'research', '')
  comm = Communicate('research')
  window = Form(comm, {'_projectID': '', 'type': ['x0']})
  qtbot.addWidget(window)
  while comm.backendThread.worker.backend is None:
    qtbot.wait(100)
  for idx, (key, _) in enumerate(window.allUserElements):
    elementName = f"key_{idx}"
    if key=='name':
      getattr(window, elementName).setText('A testing project')
  window.saveBtn.click()

  # add a comment
  qtbot.wait(1000)  # wait for backend to finish
  df = comm.backendThread.worker.backend.db.getView(f'viewDocType/x0')               # easy method for testing
  projID = df[df['name']=='A testing project']['id'].values[0]
  doc = {'comment':'# PASTA-ELN | The favorite ELN for experimental scientists\n - This is an example .eln output for sharing between different ELNs.',
         'id': projID}
  comm.uiRequestTask.emit(Task.EDIT_DOC, {'doc':doc, 'newProjID':''})

  # test scanning
  qtbot.wait(1000)  # wait for backend to finish
  loop = QEventLoop()# now use event loop to wait for backend to finish, nicer if using functions
  @Slot(dict)
  def getDoc(doc):
    dirName = comm.basePath/doc['branch'][0]['path']
    shutil.copy(Path(__file__).parent.parent/'pasta_eln'/'Resources'/'Icons'/'pasta512.png', dirName)
    loop.quit()
  comm.backendThread.worker.beSendDoc.connect(getDoc)
  comm.uiRequestDoc.emit(projID)
  loop.exec()
  @Slot(str)
  def didScan(_):
    loop.quit()
  comm.backendThread.worker.beSendTaskReport.connect(didScan)
  comm.uiRequestTask.emit(Task.SCAN, {'docID':projID})
  loop.exec()

  # move to top: directly in the database: not sure why
  cmd = "UPDATE branches SET child='-1' WHERE path=='PastasExampleProject/pasta512.png'"
  comm.uiSendSQL.emit([{'type':'one', 'cmd':cmd}])

  qtbot.wait(1000)  # wait for backend to finish
  verify(comm, projID, 0)
  print(f'{"*"*40}\nEND TEST 10\n{"*"*40}')
  comm.shutdownBackendThread()

  errors = [record for record in caplog.records if record.levelno >= logging.ERROR]
  assert not errors, f"Logging errors found: {[record.getMessage() for record in errors]}"
  return



""" widget that shows the table of the items """
import itertools
import logging
from enum import Enum
from pathlib import Path
from typing import Any
import pandas as pd
import qtawesome as qta
from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import QDialog, QFileDialog, QTableWidget, QTableWidgetItem, QVBoxLayout
from ...miscTools import callAddOn
from ..guiCommunicate import Communicate
from ..guiStyle import TextButton, space, widgetAndLayout
from .key_delegate import KeyDelegate
from .link_online_delegate import LinkOnlineDelegate
from .lookup_delegate import LookupDelegate

COLUMN_NAMES = ['key','label','PURL','', '']
COLUMN_WIDTH = [200,  400,   250, 50, 50]


class Editor(QDialog):
  """ widget that shows the table of the items """
  def __init__(self, comm:Communicate):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
    """
    super().__init__()
    self.comm = comm
    self.comm.backendThread.worker.beSendSQL.connect(self.onGetData)
    self.data:pd.DataFrame = pd.DataFrame()
    self.df0:pd.DataFrame = pd.DataFrame()
    self.df1:pd.DataFrame = pd.DataFrame()
    self.setMinimumWidth(1000)
    self.setWindowTitle('Edit definitions')

    ### GUI elements
    mainL = QVBoxLayout()
    mainL.setSpacing(space['l'])
    self.setLayout(mainL)
    ### Table
    self.table = QTableWidget(1, 5)
    self.table.verticalHeader().hide()
    self.table.setAlternatingRowColors(True)
    self.table.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
    self.table.setHorizontalHeaderLabels(COLUMN_NAMES)
    for idx, width in enumerate(COLUMN_WIDTH):
      self.table.setColumnWidth(idx, width)
    self.keyDelegate = KeyDelegate()
    self.table.setItemDelegateForColumn(0, self.keyDelegate)
    self.linkOnlineDelegate = LinkOnlineDelegate()
    self.table.setItemDelegateForColumn(3, self.linkOnlineDelegate)
    self.lookupDelegate     = LookupDelegate()
    self.table.setItemDelegateForColumn(4, self.lookupDelegate)
    self.table.horizontalHeader().setStretchLastSection(True)
    mainL.addWidget(self.table)
    ### final button box
    _, buttonLineL = widgetAndLayout('H', mainL, 'm')
    TextButton('Import', self, [Command.Import], buttonLineL, 'Import from Excel')
    TextButton('Export', self, [Command.Export], buttonLineL, 'Export to Excel')
    buttonLineL.addStretch(1)
    projectGroup = self.comm.configuration['projectGroups'][self.comm.projectGroup]
    if 'definition' in projectGroup.get('addOns',{}) and projectGroup['addOns']['definition']:
      TextButton('Autofill PURL',  self, [Command.AddOn], buttonLineL, 'Autofill by using add-on')
      buttonLineL.addStretch(1)
    self.saveBtn = TextButton('Save', self, [Command.Save], buttonLineL, 'Save changes')
    self.saveBtn.setShortcut('Ctrl+Return')
    TextButton('Cancel', self, [Command.Cancel],   buttonLineL, 'Discard changes')
    ### Data
    self.comm.uiSendSQL.emit([{'type':'get_df','cmd':'SELECT docType, PURL, title FROM docTypes'},
                              {'type':'get_df','cmd':'SELECT * FROM definitions'}])
    self.paint()


  @Slot(str, pd.DataFrame)
  def onGetData(self, cmd:str, data:pd.DataFrame) -> None:
    """ Handle data received from backend worker
    Args:
      cmd (str): command that was sent
      data (pd.DataFrame): DataFrame containing the data
    """
    if cmd == 'SELECT * FROM definitions':
      data = data[data['key'].str.contains(r'\.')]  #filter out rows that do not contain a . in the key column
      data['defType'] = 'attribute'
      self.df1 = data
    elif cmd == 'SELECT docType, PURL, title FROM docTypes':
      data['defType'] = 'class'
      self.df0 = data.rename({'docType':'key', 'title':'long'}, axis=1)
    self.data = pd.concat([self.df0,self.df1])[['key','long','PURL','defType']]
    self.data.rename({'long':'label'}, axis=1, inplace=True)
    self.paint()


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.Export:
      fileName = QFileDialog.getSaveFileName(self, 'Save table to .csv file', str(Path.home()), '*.csv')[0]
      if fileName != '':
        self.getDataframe().to_csv(fileName, index=False)
    elif command[0] is Command.Import:
      fileName = QFileDialog.getOpenFileName(self, 'Read table from .csv file', str(Path.home()), '*.csv')[0]
      if fileName != '':
        self.data = pd.read_csv(fileName).fillna('')
        self.paint()
    elif command[0] is Command.AddOn:
      try:
        self.data = callAddOn('definition_autofill', self.comm, self.data, self)
        self.paint()
      except Exception:
        pass
    elif command[0] is Command.Cancel:
      self.reject()
    elif command[0] is Command.Save:
      tasks:list[dict[str,Any]] = []
      for _, row in self.getDataframe().iterrows():
        key, description, purl, dType = row.values
        if dType == 'class':
          tasks.append({'type':'one',
                   'cmd':f"UPDATE docTypes SET PURL='{purl}', title='{description}' WHERE docType = '{key}'"})
        else:
          tasks.append({'type':'one', 'cmd':'INSERT OR REPLACE INTO definitions VALUES (?, ?, ?);',
                        'list':[key, description, purl]})
      self.comm.uiSendSQL.emit(tasks)
      self.accept()
    else:
      logging.error('Command unknown: %s',command, exc_info=True)
    return


  def paint(self) -> None:
    """ Show data frame in the GUI """
    self.table.setRowCount(len(self.data))
    nRows, nCols = self.data.shape
    for i, j in itertools.product(range(nRows), range(nCols-1)):
      rowType = self.data.iloc[i, 3]
      icon = qta.icon('msc.symbol-class' if rowType=='class' else 'msc.symbol-property', scale_factor=1)
      item = QTableWidgetItem(self.data.iloc[i, j])
      if j==0:
        item.setIcon(icon)
      self.table.setItem(i, j, item)
    return


  def getDataframe(self) -> pd.DataFrame:
    """ Get dataframe from table """
    model = self.table.model()
    data = []
    for row in range(model.rowCount()):
      rowRes = [model.index(row, column).data() for column in range(3)]
      data.append(rowRes)
    df = pd.DataFrame(data)
    df = df.rename({0:'key',1:'description',2:'PURL'}, axis=1)
    df = df.merge(self.data, how='left', left_on='key', right_on='key')
    df = df.drop(['label','PURL_y'], axis=1).rename({'PURL_x':'PURL'}, axis=1)
    # df['defType']=self.data['defType'].values    # ignore
    return df


  def reject(self) -> None:
    """ Reject the dialog, stop the thread and disconnect signals """
    self.comm.backendThread.worker.beSendSQL.disconnect(self.onGetData)
    super().reject()


  def accept(self) -> None:
    """ Accept the dialog, stop the thread and disconnect signals """
    self.comm.backendThread.worker.beSendSQL.disconnect(self.onGetData)
    super().accept()


class Command(Enum):
  """ Commands used in this file """
  Save   = 1
  Cancel = 2
  Import = 3
  Export = 4
  AddOn  = 5



#!/usr/bin/python3
from pasta_eln.UI.guiCommunicate import Communicate
from pasta_eln.UI.project import Project
from .test_34_GUI_Form import getTable

import logging, os, shutil, tempfile
from urllib import request
from pasta_eln.backendWorker.backend import Backend
from pasta_eln.backendWorker.worker import Task
from pasta_eln.backendWorker.inputOutput import importELN
from pasta_eln.miscTools import getConfiguration


def test_simple(qtbot, caplog):

  # remove old data
  configuration, _ = getConfiguration('research')
  backend = Backend(configuration, 'research')
  dirName = backend.basePath
  backend.exit()
  try:
    shutil.rmtree(dirName)
    os.makedirs(dirName)
  except Exception:
    pass

  comm = Communicate('research')
  while comm.backendThread.worker.backend is None:
    qtbot.wait(100)
  window = Project(comm)
  window.setMinimumSize(1024,800)
  window.show()

  # location of data and temp
  tempDir = tempfile.gettempdir()
  allELNs = {'PASTA_ELN.eln'      : '', #no download, created by test_11_Export.py
             'Pasta.eln'          :'PASTA/PASTA.eln',
             'elabFTW.eln'        :'elabftw/export.eln',
             'SampleDB.eln'       :'SampleDB/sampledb_export.eln',
             'kadi4mat_1.eln'     :'kadi4mat/collections-example.eln',
             'kadi4mat_2.eln'     :'kadi4mat/records-example.eln',
             'openSemanticLab.eln':'OpenSemanticLab/MinimalExample.osl.eln'
            }
  urlBase = 'https://github.com/TheELNConsortium/TheELNFileFormat/raw/refs/heads/master/examples/'

  for eln, urlSubPath in allELNs.items():
    elnFileName = f'{tempDir}/{eln}'
    if urlSubPath:
      request.urlretrieve(f'{urlBase}{urlSubPath}', elnFileName)
    print(f'\n\n{"="*30}\nStart with {eln} on {elnFileName}')
    projName = f'{eln[:-4]} Import'
    comm.uiRequestTask.emit(Task.ADD_DOC, {'docType':'x0', 'doc':{'name':projName}, 'hierStack':[]})
    df = getTable(qtbot, comm, 'x0')
    projID = df[df['name']==projName]['id'].values[0]
    comm.uiRequestTask.emit(Task.IMPORT_ELN, {'fileName':elnFileName, 'projID':projID})
    qtbot.wait(2000)  # wait until import is done

    comm.changeProject.emit(projID, '')
    qtbot.addWidget(window)
    qtbot.wait(1000)
    path = qtbot.screenshot(window)
    print(path)



  comm.shutdownBackendThread()

  errors = [record for record in caplog.records if record.levelno >= logging.ERROR]
  assert not errors, f"Logging errors found: {[record.getMessage() for record in errors]}"



  #   reply, statistics = importELN(backend, elnFileName, projID)
  #   print(reply,'\n',statistics)
  #   assert statistics['num. files']<=statistics['types'].get('File',0) ,'Files do not make sense'
  #   #statistics type File can be larger than num.Files because there are remote files which are not Files but in the @type
  #   print(backend.checkDB(minimal=True))

  return



""" widget that shows the table and the details of the items """
from typing import Any
from PySide6.QtCore import Slot
from PySide6.QtWidgets import QSplitter, QVBoxLayout, QWidget
from .details import Details
from .guiCommunicate import Communicate
from .table import Table


class DocTypes(QWidget):
  """ widget that shows the table and the details of the items """
  def __init__(self, comm:Communicate):
    super().__init__()
    comm.changeTable.connect(self.changeTable)
    comm.changeDetails.connect(self.changeDetails)
    # GUI elements
    table = Table(comm)
    self.details = Details(comm)
    self.details.resizeEvent = self.resizeWidget                                                # type: ignore
    splitter = QSplitter()
    splitter.setHandleWidth(10)
    splitter.addWidget(table)
    splitter.addWidget(self.details)
    splitter.setContentsMargins(0,0,0,0)
    splitter.setSizes([1800, 600])                                           # will be scaled down accordingly
    mainL = QVBoxLayout()
    mainL.setSpacing(0)
    mainL.setContentsMargins(0, 0, 0, 0)
    mainL.addWidget(splitter)
    self.setLayout(mainL)


  @Slot(str, str)
  def changeTable(self) -> None:                                             # pylint: disable=unused-argument
    """What happens when user clicks to change doc-type -> show table """
    self.details.hide()
    return


  @Slot(str)
  def changeDetails(self, docID:str) -> None:
    """
    What happens when user clicks to change details
    -> show show details

    Args:
      docID (str): document ID
    """
    if docID!='':
      self.details.show()
    return


  def resizeWidget(self, _:Any) -> None:
    """ called if splitter resizes details-view  """
    self.details.resizeWidth(self.details.width())
    return
